#include "5ro14_husky_controller/MyController.hpp"
#include <cmath>

#include "geometry_msgs/Twist.h"
#include "visualization_msgs/Marker.h"

#define _USE_MATH_DEFINES

namespace husky_controller
{
  MyController::MyController(ros::NodeHandle &nodeHandle) :
      m_nodeHandle(&nodeHandle)
  {
    std::string scanTopicName;
    if (!m_nodeHandle->getParam("scan_topic_name", scanTopicName))
    {
      ROS_ERROR("Load scan_topic_name param fail!!");
      return;
    }

    int scanTopicQueueSize;
    if (!m_nodeHandle->getParam("scan_topic_queue_size", scanTopicQueueSize))
    {
      ROS_ERROR("Load scan_topic_queue_size param fail!!");
      return;
    }

    m_pGainAngle = 1.0;
    if (!m_nodeHandle->getParam("p_gain_angle", m_pGainAngle))
    {
      ROS_WARN("Load p_gain_angle param fail!! Use the default value %f", m_pGainAngle);
    }

    m_laserScanSubscriber = m_nodeHandle->subscribe(scanTopicName, scanTopicQueueSize, &MyController::laserCallBack, this);

    m_commandVelocityPublisher = m_nodeHandle->advertise<geometry_msgs::Twist>("/cmd_vel", 1);

    m_markerPillarPublisher = m_nodeHandle->advertise<visualization_msgs::Marker>("/pillar_marker", 100);
  }

  
  void MyController::laserCallBack(const sensor_msgs::LaserScan::ConstPtr &msg)
  {
    float minVal = msg->range_max;
    unsigned int directionId;

    for (unsigned int i = 0; i < msg->ranges.size(); ++i)
    {
      if (std::isnormal(msg->ranges[i]))
      {
        if (minVal > msg->ranges[i])
        {
          minVal = msg->ranges[i];
          directionId = i;
        }
      }
    }

    //ROS_INFO_STREAM(minVal);

    // Compute distance and angle of the pillar from the laser scan with respect to the robot.
    double pillarAngle = msg->angle_min + msg->angle_increment * directionId;
    double pillarDist = minVal;

    geometry_msgs::Twist control_vel;
    double angleError = -pillarAngle;

    if (pillarDist < msg->range_max)
    {
      control_vel.angular.z = m_pGainAngle * angleError;
      control_vel.linear.x = 2;
    }
    else // dont see the pillar => turn
    {
      control_vel.angular.z = m_pGainAngle * M_PI / 2;
      control_vel.linear.x = 0;
    }

    // Publish on /cmd_vel
    m_commandVelocityPublisher.publish(control_vel);

    // Construct a marker
    visualization_msgs::Marker pillarMarker;
    // Here msg->header.frame_id = "base_laser"
    pillarMarker.header.frame_id = msg->header.frame_id;
    pillarMarker.header.stamp = ros::Time();
    pillarMarker.ns = "marker";
    pillarMarker.id = 1;
    pillarMarker.type = visualization_msgs::Marker::SPHERE;
    pillarMarker.action = visualization_msgs::Marker::ADD;
    pillarMarker.pose.position.x = pillarDist * std::cos(pillarAngle);
    pillarMarker.pose.position.y = pillarDist * std::sin(pillarAngle);
    pillarMarker.pose.position.z = 0.0;
    pillarMarker.pose.orientation.x = 0.0;
    pillarMarker.pose.orientation.y = 0.0;
    pillarMarker.pose.orientation.z = 0.0;
    pillarMarker.pose.orientation.w = 1.0;
    pillarMarker.scale.x = 0.4;
    pillarMarker.scale.y = 0.4;
    pillarMarker.scale.z = 1;
    pillarMarker.color.a = 1.0;
    pillarMarker.color.r = 0.0;
    pillarMarker.color.g = 0.0;
    pillarMarker.color.b = 1.0;

    // publish the marker
    m_markerPillarPublisher.publish(pillarMarker);
  }
} /* namespace */
